/* 
  $RCSfile: CardOSAPI_Token.h,v $

  Copyright (c) 2011 Atos IT Solutions and Services
  www.atos.net/cardos
  All rights reserved.
  
  C++ interface header file for the CardOS API smart card middleware.
  version 1.3, October 10, 2012. Subject to change without notice.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions 
  are met:

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Atos nor the names of its contributors may be
      used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.

*/

#ifndef CARDOSAPI_TOKEN_H
#define CARDOSAPI_TOKEN_H

#if defined (WINDOWS) || defined (_WINDOWS)
  // Disable warnings about C++ exception specfications ignored by VC++
# pragma warning(disable : 4290)
# include <winscard.h>
#else
# include <PCSC/winscard.h>
# include <PCSC/pcsclite.h>
#endif

#include <string>
#include <vector>
#include <list>

using namespace std;

// Error codes.
// The possible error codes and their semantics are according to
// "Common Security: CDSA and CSSM, Version 2", section 7.5.

#define CARDOS_API_OK 0

// Common error codes
#define CARDOS_API_E_INTERNAL_ERROR         (unsigned long)0x00000001
#define CARDOS_API_E_MEMORY_ERROR           (unsigned long)0x00000002
#define CARDOS_API_E_INVALID_INPUT_POINTER  (unsigned long)0x00000003
#define CARDOS_API_E_FUNCTION_FAILED        (unsigned long)0x00000004

// ACL error codes
#define CARDOS_API_E_OPERATION_AUTH_DENIED  (unsigned long)0x00000101
#define CARDOS_API_E_INVALID_SAMPLE_VALUE   (unsigned long)0x00000102

// Common error codes for specific data types
#define CARDOS_API_E_INVALID_CONTEXT_HANDLE (unsigned long)0x00000201

// CSP error codes
#define CARDOS_API_E_INVALID_KEY            (unsigned long)0x00000301
#define CARDOS_API_E_INVALID_KEY_REFERENCE  (unsigned long)0x00000302
#define CARDOS_API_E_ALGID_MISMATCH         (unsigned long)0x00000303
#define CARDOS_API_E_INVALID_KEY_LABEL      (unsigned long)0x00000304
#define CARDOS_API_E_NOT_LOGGED_IN          (unsigned long)0x00000305
#define CARDOS_API_E_BLOCK_SIZE_MISMATCH    (unsigned long)0x00000306


namespace net_atos_cardos_api {

  typedef vector<unsigned char> BinaryObject;
  
  /**
   * \class CardOSAPI_Exception
   */
  class CardOSAPI_Exception  {
  private:
    unsigned long errorCode;
    
  public:
    // Default constructor
    CardOSAPI_Exception() throw() {errorCode = CARDOS_API_E_INTERNAL_ERROR;};
    
    // Custom constructor
    CardOSAPI_Exception(unsigned long errorCode) throw() { this->errorCode = errorCode; };

    // Copy constructor
    CardOSAPI_Exception(const CardOSAPI_Exception& ex) { this->errorCode = ex.errorCode; };
    
    // Destructor
    virtual ~CardOSAPI_Exception() throw() {};
    
    virtual unsigned long getErrorCode() const {return errorCode;};
  };
    

  typedef enum {
    PADDING_NONE  = 0,
    PADDING_PKCS_V1_5 = 1
  } CardOSAPI_Padding;
  
  typedef enum {
    DIGESTINFO_NONE    = 0,
    DIGESTINFO_MD2     = 1,
    DIGESTINFO_MD5     = 2,
    DIGESTINFO_RIPE160 = 3,
    DIGESTINFO_SHA1    = 4,
    DIGESTINFO_SHA224  = 5,
    DIGESTINFO_SHA256  = 6,
    DIGESTINFO_SHA384  = 7,
    DIGESTINFO_SHA512  = 8
  } CardOSAPI_DigestInfo;

  /**
   * \class CardOSAPI_X509Identity
   * \brief 
   */
  class CardOSAPI_X509Identity
  {  
  public:
    // Default constructor.
    CardOSAPI_X509Identity() throw() : pInstance(NULL) {};
     
    // Constructor that swallows an instance of a CardOSAPI_X509Identity 
    // subclass implementation.
    CardOSAPI_X509Identity(CardOSAPI_X509Identity * pIdentity) throw() : pInstance(pIdentity) {};
    
    // Destructor.
    virtual ~CardOSAPI_X509Identity() throw() { };
    //virtual CardOSAPI_X509Identity& operator = (const CardOSAPI_X509Identity& other) throw() { return *this; };
    

    // Accessors

    // Get human readable description for the identity.
    virtual const string getDescription() const throw(CardOSAPI_Exception) { if (pInstance) return pInstance->getDescription(); throw(CardOSAPI_Exception()); };
    // Get certifiate value in ASN.1 DER coding.
    virtual BinaryObject getCertificate() const throw(CardOSAPI_Exception) { if (pInstance) return pInstance->getCertificate(); throw(CardOSAPI_Exception()); };
    // Get key length in bits.
    virtual unsigned long getKeyLength() const throw() { if (pInstance) return pInstance->getKeyLength(); else return 0; };
    // Return an identifier for the authentication object protecting the 
    // identity.
    virtual int getAuthId() const throw(CardOSAPI_Exception) { if (pInstance) return pInstance->getAuthId(); else return 0; };
    
    // Properties

    // Return whether the identity contains a private key that can be used to
    // decrypt encrypted data.
    virtual bool canDecrypt() const throw() { if (pInstance) return pInstance->canDecrypt(); else return false; };
    // Return whether the identity contains a private key that can be used to
    // compute a digital signature.
    virtual bool canSign() const throw() { if (pInstance) return pInstance->canSign(); else return false; };
    // Return whether the identity contains a private key that can be used to
    // decrypt wrapped key data.
    virtual bool canUnwrap() const throw() { if (pInstance) return pInstance->canUnwrap(); else return false; };


    // Actions

    // Compute digital signature.
    virtual BinaryObject generateSignature(CardOSAPI_Padding padding, CardOSAPI_DigestInfo digestInfo, const BinaryObject& inputData) throw(CardOSAPI_Exception)
    { 
      if (pInstance) return pInstance->generateSignature(padding, digestInfo, inputData); 
      throw(CardOSAPI_Exception());
    }

    // Unwrap a wrapped secret (symmetric) key.
    virtual BinaryObject unwrapKey(const BinaryObject &wrappedKey) throw(CardOSAPI_Exception)
    {
      if (pInstance) return pInstance->unwrapKey(wrappedKey); 
      throw(CardOSAPI_Exception());
    }

  private:
    CardOSAPI_X509Identity * pInstance;
  };

  typedef enum {
    PIN_NOT_AUTHENTICATED,
    PIN_AUTHENTICATED,
    PIN_BLOCKED
  } CardOSAPI_PinState;

 typedef list<CardOSAPI_X509Identity> X509IdentityList;

  /**
   * \class CardOSAPI_Token
   * \brief 
   */
  class CardOSAPI_Token
  {
  public:
    virtual ~CardOSAPI_Token() {};
    
    virtual void log(const std::string& msg) = 0;

    virtual unsigned int probe(string & tokenUid) throw(CardOSAPI_Exception) = 0;
    virtual void establish(const char *guid, const string cacheDirectory, const string workDirectory) throw(CardOSAPI_Exception) = 0;
    
    virtual string getFriendlyName() const throw() = 0;
    virtual string getHWSerialNumber() const throw() = 0;
    
    // Return an identifier for the authentication object used to authenticate
    // the user against the token.
    virtual int getUserAuthId() const throw(CardOSAPI_Exception) = 0;
    // Get verification state of PIN identified by pinNum.
    virtual CardOSAPI_PinState getPinState(int pinNum) throw() = 0;
    // Change PIN identified by pinNum.
    virtual void changePIN(int pinNum, const unsigned char *oldPin, size_t oldPinLength, const unsigned char *newPin, size_t newPinLength) throw(CardOSAPI_Exception) = 0;
    // Verify PIN identified by pinNum.
    virtual void verifyPIN(int pinNum, const unsigned char *pin, size_t pinLength) throw(CardOSAPI_Exception) = 0;
    // Unverify PIN identified by pinNum.
    virtual void unverifyPIN(int pinNum) throw(CardOSAPI_Exception) = 0;
    
    // Get list of X.509 identities stored on the token.
    virtual X509IdentityList& getX509Identities() throw() = 0;
    
  protected:
    CardOSAPI_Token() {};
  };
  

/**
 * \class CardOSAPI_TokenFactory
 * \brief 
 */
  
  class CardOSAPI_TokenFactory
  {
    public:   
      static CardOSAPI_Token * create(const SCARD_READERSTATE * pReaderState);
      static void destroy(CardOSAPI_Token * pToken);
  };
  
} /* namespace net_atos_cardos_api */


#endif /* CARDOSAPI_TOKEN_H  */

