/*
 *  CardOSKeyHandle.cpp
 *
 *  Copyright (c) 2011 Atos IT Solutions and Services GmbH
 *
 *  CardOS TokenD implementation based on XYZKeyHandle.cpp provided 
 *  as part of the Apple TokenD developer information available from 
 *  dts@apple.com .
 *
 *  Use diff -c XYZKeyHandle.cpp CardOSKeyHandle.cpp to generate a patch.
 *
 *  See below for the copyright notice of the original file
 *  and the license applying to both the original and modified code.
 *
 *  Copyright (c) 2004 Apple Computer, Inc. All Rights Reserved.
 * 
 *  @APPLE_LICENSE_HEADER_START@
 *  
 *  This file contains Original Code and/or Modifications of Original Code
 *  as defined in and that are subject to the Apple Public Source License
 *  Version 2.0 (the 'License'). You may not use this file except in
 *  compliance with the License. Please obtain a copy of the License at
 *  http://www.opensource.apple.com/apsl/ and read it before using this
 *  file.
 *  
 *  The Original Code and all software distributed under the License are
 *  distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 *  EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 *  INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 *  Please see the License for the specific language governing rights and
 *  limitations under the License.
 *  
 *  @APPLE_LICENSE_HEADER_END@
 */

#include <security_utilities/debugging.h>
#include <security_utilities/utilities.h>
#include <security_cdsa_utilities/cssmerrors.h>
#include <Security/cssmerr.h>

#include "CardOSKeyHandle.h"
#include "CardOSRecord.h"
#include "CardOSToken.h"
#include "CardOSError.h"

using namespace Tokend;


//
// CardOSKeyHandle
//
CardOSKeyHandle::CardOSKeyHandle(CardOSToken &cardOStoken,
                                 const Tokend::MetaRecord &metaRecord, CardOSKeyRecord &cardOSKey) :
  Tokend::KeyHandle(metaRecord, &cardOSKey), mToken(cardOStoken), mKey(cardOSKey)
{ 
  mToken.cardosToken->log("CardOSKeyHandle::CardOSKeyHandle()");
}


CardOSKeyHandle::~CardOSKeyHandle()
{ 
  mToken.cardosToken->log("CardOSKeyHandle::~CardOSKeyHandle()");
}


void CardOSKeyHandle::getKeySize(CSSM_KEY_SIZE &keySize)
{
  secdebug("crypto", "CardOSKeyHandle::getKeySize");
  mToken.cardosToken->log("CardOSKeyHandle::getKeySize");
  
  CssmError::throwMe(CSSM_ERRCODE_FUNCTION_NOT_IMPLEMENTED);
}


uint32 CardOSKeyHandle::getOutputSize(const Context &context,
                                      uint32 inputSize, bool encrypting)
{
  secdebug("crypto", "CardOSKeyHandle::getOutputSize");
  mToken.cardosToken->log("CardOSKeyHandle::getOutputSize");
  
  CssmError::throwMe(CSSM_ERRCODE_FUNCTION_NOT_IMPLEMENTED);
}


void CardOSKeyHandle::generateSignature(const Context &context,
                                        CSSM_ALGORITHMS signOnly,
                                        const CssmData &input, 
                                        CssmData &signature)
{
  secdebug("crypto", "CardOSKeyHandle::generateSignature alg: %u signOnly: %u", context.algorithm(), signOnly);
  IFDUMPING("crypto", context.dump("signature context"));
  
  mToken.cardosToken->log("CardOSKeyHandle::generateSignature");
  
  if (context.type() != CSSM_ALGCLASS_SIGNATURE) {
    CssmError::throwMe(CSSMERR_CSP_INVALID_CONTEXT);
  }
  
  if (context.algorithm() != CSSM_ALGID_RSA) {
    CssmError::throwMe(CSSMERR_CSP_INVALID_ALGORITHM);
  }
  
  // Find out which digest algorithm applies.
  CardOSAPI_DigestInfo cardosDI = DIGESTINFO_NONE;
  switch (signOnly) {
	case CSSM_ALGID_NONE:
	  cardosDI = DIGESTINFO_NONE;
	  break;
	case CSSM_ALGID_MD5:
	  cardosDI = DIGESTINFO_MD5;
	  break;
    case CSSM_ALGID_SHA1:
      cardosDI = DIGESTINFO_SHA1;
	  break;
    case CSSM_ALGID_SHA224:
      cardosDI = DIGESTINFO_SHA224;
	  break;
    case CSSM_ALGID_SHA256:
      cardosDI = DIGESTINFO_SHA256;
	  break;
    case CSSM_ALGID_SHA384:
      cardosDI = DIGESTINFO_SHA384;
	  break;
    case CSSM_ALGID_SHA512:
      cardosDI = DIGESTINFO_SHA512;
	  break;
    default:
      CssmError::throwMe(CSSMERR_CSP_INVALID_DIGEST_ALGORITHM);
      break;
  }
  
  // Get padding; default to PKCS#1 style padding.
  uint32 cssmPadding = CSSM_PADDING_PKCS1;
  context.getInt(CSSM_ATTRIBUTE_PADDING, cssmPadding);
  
  CardOSAPI_Padding cardosPadding = PADDING_NONE;
  switch (cssmPadding) {
    case CSSM_PADDING_PKCS1:
      cardosPadding = PADDING_PKCS_V1_5;
      break;
    case CSSM_PADDING_NONE:
      cardosPadding = PADDING_NONE;
      break;
    default:
      CssmError::throwMe(CSSMERR_CSP_INVALID_ATTR_PADDING);
      break;
  }
  
  // Convert CssmData input to BinaryObject.
  BinaryObject tbs(input.Length);
  memcpy(&tbs[0], input.Data, input.Length);
  
  BinaryObject sig;
  
  // Call generate Signature
  try {
	sig = mKey.getRecord().generateSignature(cardosPadding, cardosDI , tbs);
	
	signature.Data = reinterpret_cast<unsigned char *>(mToken.support()->malloc(sig.size()));
	if (signature.Data == NULL) {
	  CssmError::throwMe(CSSM_ERRCODE_MEMORY_ERROR);
	}
    memcpy(signature.Data, &sig[0], sig.size());
    signature.Length = sig.size();
  }
  catch(CardOSAPI_Exception cardosEx) {
    CardOSError::throwMe(cardosEx);
  }
}



void CardOSKeyHandle::unwrapKey(const Context &context,
                                const AccessCredentials *cred, const AclEntryPrototype *access,
                                const CssmKey &wrappedKey, CSSM_KEYUSE usage,
                                CSSM_KEYATTR_FLAGS attributes, CssmData *descriptiveData,
                                CSSM_HANDLE &hndUnwrappedKey, CssmKey &unwrappedKey)
{
  secdebug("crypto", "CardOSKeyHandle::unwrapKey alg: %u", context.algorithm());  
  IFDUMPING("crypto", context.dump("unwrapKey context"));
  
  mToken.cardosToken->log("CardOSKeyHandle::unwrapKey");
  
  if (context.type() != CSSM_ALGCLASS_ASYMMETRIC) {
    CssmError::throwMe(CSSMERR_CSP_INVALID_CONTEXT);
  }
  
  // validate wrappedKey
  if (wrappedKey.keyClass() != CSSM_KEYCLASS_SESSION_KEY) {
    CssmError::throwMe(CSSMERR_CSP_INVALID_KEY_CLASS);
  }
  
  if (wrappedKey.blobType() != CSSM_KEYBLOB_WRAPPED) {
    CssmError::throwMe(CSSMERR_CSP_KEY_BLOB_TYPE_INCORRECT);
  }
  
  // validate requested storage and usage
  if (   !(attributes & CSSM_KEYATTR_RETURN_DATA) 
	  || (attributes & (CSSM_KEYATTR_RETURN_REF | CSSM_KEYATTR_RETURN_NONE 
						| CSSM_KEYATTR_PERMANENT | CSSM_KEYATTR_PRIVATE)) != 0) {
    CssmError::throwMe(CSSMERR_CSP_INVALID_KEYATTR_MASK);
  }
  
  // prepare outgoing header
  CssmKey::Header &header = unwrappedKey.header();
  header.clearPod();
  header.HeaderVersion = CSSM_KEYHEADER_VERSION;
  header.cspGuid(gGuidAppleSdCSPDL);
  header.blobType(CSSM_KEYBLOB_RAW);
  header.algorithm(wrappedKey.algorithm());
  header.keyClass(wrappedKey.keyClass());
  header.KeyUsage = usage;
  header.KeyAttr = attributes & ~(CSSM_KEYATTR_RETURN_DATA | CSSM_KEYATTR_RETURN_REF | CSSM_KEYATTR_RETURN_NONE);
  
  // defaults (change as needed)
  header.StartDate = wrappedKey.header().StartDate;
  header.EndDate = wrappedKey.header().EndDate;
  unwrappedKey.KeyData.Data = NULL;	// ignore possible incoming KeyData
  unwrappedKey.KeyData.Length = 0;
  
  // validate wrappedKey format
  if (wrappedKey.blobFormat() != CSSM_KEYBLOB_WRAPPED_FORMAT_PKCS7 ) {
    CssmError::throwMe( CSSMERR_CSP_INVALID_ATTR_WRAPPED_KEY_FORMAT );
  }
  
  // There is no descriptiveData in a PKCS7 wrapped blob.
  if (descriptiveData) {
    descriptiveData->Data = NULL;
    descriptiveData->Length = 0;
  }
  //  const AccessCredentials *keyCred = context.get<const AccessCredentials>( CSSM_ATTRIBUTE_ACCESS_CREDENTIALS );
  BinaryObject wrappedKeyData(wrappedKey.KeyData.Length);
  memcpy(&(wrappedKeyData[0]), wrappedKey.KeyData.Data, wrappedKey.KeyData.Length);
  
  // Call unwrap key
  try {
	BinaryObject unwrappedKeyData = mKey.getRecord().unwrapKey(wrappedKeyData);
	unwrappedKey.KeyData.Data = reinterpret_cast<unsigned char *>(mToken.support()->malloc(unwrappedKeyData.size()));
	if (unwrappedKey.KeyData.Data == NULL) {
	  CssmError::throwMe(CSSM_ERRCODE_MEMORY_ERROR);
	}
    memcpy(unwrappedKey.KeyData.Data, &unwrappedKeyData[0], unwrappedKeyData.size());
    unwrappedKey.KeyData.Length = unwrappedKeyData.size();
  }
  catch(CardOSAPI_Exception cardosEx) {
    CardOSError::throwMe(cardosEx);
  }
    
  // Assuming a CSSM_KEYBLOB_RAW_FORMAT_OCTET_STRING from here on.
  header.blobFormat(CSSM_KEYBLOB_RAW_FORMAT_OCTET_STRING);
  header.LogicalKeySizeInBits = unwrappedKey.length() * 8;  
  
}


void CardOSKeyHandle::verifySignature(const Context &context,
                                      CSSM_ALGORITHMS signOnly, const CssmData &input, const CssmData &signature)
{
  secdebug("crypto", "CardOSKeyHandle::verifySignature");
  
  CssmError::throwMe(CSSM_ERRCODE_FUNCTION_NOT_IMPLEMENTED);
}


void CardOSKeyHandle::generateMac(const Context &context,
                                  const CssmData &input, CssmData &output)
{
  secdebug("crypto", "CardOSKeyHandle::generateMac");
  
  CssmError::throwMe(CSSM_ERRCODE_FUNCTION_NOT_IMPLEMENTED);
}


void CardOSKeyHandle::verifyMac(const Context &context,
                                const CssmData &input, const CssmData &compare)
{
  secdebug("crypto", "CardOSKeyHandle::verifyMac");
  CssmError::throwMe(CSSM_ERRCODE_FUNCTION_NOT_IMPLEMENTED);
}


void CardOSKeyHandle::encrypt(const Context &context,
                              const CssmData &clear, CssmData &cipher)
{
  secdebug("crypto", "CardOSKeyHandle::encrypt");
  CssmError::throwMe(CSSM_ERRCODE_FUNCTION_NOT_IMPLEMENTED);
}


void CardOSKeyHandle::decrypt(const Context &context,
                              const CssmData &cipher, CssmData &clear)
{
  secdebug("crypto", "CardOSKeyHandle::decrypt alg: %u", context.algorithm());
  CssmError::throwMe(CSSMERR_CSP_KEY_USAGE_INCORRECT);
}


void CardOSKeyHandle::exportKey(const Context &context,
                                const AccessCredentials *cred, CssmKey &wrappedKey)
{
  secdebug("crypto", "CardOSKeyHandle::exportKey");
  CssmError::throwMe(CSSM_ERRCODE_FUNCTION_NOT_IMPLEMENTED);
}


//
// CardOSKeyHandleFactory
//
CardOSKeyHandleFactory::~CardOSKeyHandleFactory()
{
}


Tokend::KeyHandle *CardOSKeyHandleFactory::keyHandle(Tokend::TokenContext *tokenContext, const Tokend::MetaRecord &metaRecord,
                                                     Tokend::Record &record) const
{
  CardOSKeyRecord &key = dynamic_cast<CardOSKeyRecord &>(record);
  CardOSToken &cardosToken = static_cast<CardOSToken &>(*tokenContext);
  
  cardosToken.cardosToken->log("CardOSKeyHandleFactory::keyHandle");
  return new CardOSKeyHandle(cardosToken, metaRecord, key);
}
