/*
 *  CardOSToken.cpp
 *
 *  Copyright (c) 2011 Atos IT Solutions and Services GmbH
 *
 *  CardOS TokenD implementation based on XYZToken.cpp provided 
 *  as part of the Apple TokenD developer information available from 
 *  dts@apple.com .
 *
 *  Use diff -c XYZToken.cpp CardOSToken.cpp to generate a patch.
 *
 *  See below for the copyright notice of the original file
 *  and the license applying to both the original and modified code.
 *
 *  Copyright (c) 2004 Apple Computer, Inc. All Rights Reserved.
 * 
 *  @APPLE_LICENSE_HEADER_START@
 *  
 *  This file contains Original Code and/or Modifications of Original Code
 *  as defined in and that are subject to the Apple Public Source License
 *  Version 2.0 (the 'License'). You may not use this file except in
 *  compliance with the License. Please obtain a copy of the License at
 *  http://www.opensource.apple.com/apsl/ and read it before using this
 *  file.
 *  
 *  The Original Code and all software distributed under the License are
 *  distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 *  EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 *  INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 *  Please see the License for the specific language governing rights and
 *  limitations under the License.
 *  
 *  @APPLE_LICENSE_HEADER_END@
 */


#include "CardOSToken.h"

#include "Adornment.h"
#include "AttributeCoder.h"
#include "CardOSError.h"
#include "SCardError.h"
#include <security_cdsa_client/aclclient.h>
#include <map>
#include <vector>



using CssmClient::AclFactory;
using namespace Tokend;


CardOSToken::CardOSToken()
{
  secdebug("tokend", "CardOSToken::CardOSToken()");
  
  cardosToken = NULL;
  mTokenContext = this;   
  mSchema = NULL;
}

CardOSToken::~CardOSToken()
{
  secdebug("tokend", "CardOSToken::~CardOSToken()");
  if (cardosToken) {
    cardosToken->log("CardOSToken::~CardOSToken()");
	terminate(0, 0);
  }
  
  if( mSchema != NULL ) {
    delete mSchema;
  }
}


void CardOSToken::changePIN(int pinNum,
                            const unsigned char *oldPin, size_t oldPinLength,
                            const unsigned char *newPin, size_t newPinLength)
{
  secdebug("tokend", "CardOSToken::changePIN");
  
  if (cardosToken == NULL) {
    CssmError::throwMe(CSSM_ERRCODE_SELF_CHECK_FAILED);
  }
    
  cardosToken->log("CardOSToken::changePIN");
  
#if 1
  if (pinNum != cardosToken->getUserAuthId()) {
	CssmError::throwMe(CSSM_ERRCODE_SAMPLE_VALUE_NOT_SUPPORTED);
  }
#endif
  
  try {
    cardosToken->changePIN(pinNum, oldPin, oldPinLength, newPin, newPinLength);
  }
  catch(CardOSAPI_Exception cardosEx) {
    CardOSError::throwMe(cardosEx);
  }
}


bool CardOSToken::isLocked()
{
  secdebug("tokend", "CardOSToken::isLocked");
  
  if (cardosToken == NULL) {
    CssmError::throwMe(CSSM_ERRCODE_SELF_CHECK_FAILED);
  }
  
  cardosToken->log("CardOSToken::isLocked");
  
  return (cardosToken->getPinState(1 /* ToDo */) != PIN_AUTHENTICATED);
}


void CardOSToken::verifyPIN(int pinNum, const uint8_t *pin, size_t pinLength)
{
  secdebug("tokend", "CardOSToken::verifyPIN");
  
  if (cardosToken == NULL) {
    CssmError::throwMe(CSSM_ERRCODE_SELF_CHECK_FAILED);
  }
  
  cardosToken->log("CardOSToken::verifyPIN");
  
  try {
    cardosToken->verifyPIN(pinNum, pin, pinLength);
  }
  catch(CardOSAPI_Exception cardosEx) {
    CardOSError::throwMe(cardosEx);
  }
}


void CardOSToken::unverifyPIN(int pinNum)
{
  secdebug("tokend", "CardOSToken::unverifyPIN");
  
  if (cardosToken == NULL) {
    CssmError::throwMe(CSSM_ERRCODE_SELF_CHECK_FAILED);
  }
    
  cardosToken->log("CardOSToken::unverifyPIN");
  
  try {
    cardosToken->unverifyPIN(pinNum);
  }
  catch(CardOSAPI_Exception cardosEx) {
    CardOSError::throwMe(cardosEx);
  }
}


uint32 CardOSToken::probe(SecTokendProbeFlags flags, char tokenUid[TOKEND_MAX_UID])
{
  secdebug("tokend", "CardOSToken::probe");
  
  uint32 score = 0;
  string uid;
  
  if (cardosToken == NULL) {
    cardosToken = CardOSAPI_TokenFactory::create((*startupReaderInfo)());
  }
  
  cardosToken->log("CardOSToken::probe");
  cardosToken->log((flags & kSecTokendProbeKeepToken) ? "true" : "false");
  
  try {
    score = cardosToken->probe(uid);
	memset(tokenUid, 0x00, TOKEND_MAX_UID);
	strncpy(tokenUid, uid.c_str(), MIN(uid.size(), TOKEND_MAX_UID-1));
	cardosToken->log(tokenUid);
	// ToDo
    //if (!(flags & kSecTokendProbeKeepToken)) {
      //CardOSAPI_TokenFactory::destroy(cardosToken);
      //cardosToken = NULL;
    //}
  }
  catch(CardOSAPI_Exception cardosEx) {
    CardOSError::throwMe(cardosEx);
  }
  
  return score;
}


void CardOSToken::establish(const CSSM_GUID *guid, uint32 subserviceId, SecTokendEstablishFlags flags,
                            const char *cacheDirectory, const char *workDirectory, char mdsDirectory[PATH_MAX],
                            char printName[PATH_MAX])
{
  secdebug("tokend", "CardOSToken::establish");
  
  string friendlyName;
  
  if (cardosToken == NULL) {
    cardosToken = CardOSAPI_TokenFactory::create((*startupReaderInfo)());
  }
  
  cardosToken->log("CardOSToken::establish");
  
  try {
    mGuid = *guid; /* ToDo: establish doesn't need guid as parameter */
    cardosToken->establish( NULL /*guid*/, cacheDirectory, workDirectory);
    friendlyName = cardosToken->getFriendlyName();
    memset(printName, 0x00, sizeof(printName));
    strncpy(printName, friendlyName.c_str(), MIN(PATH_MAX, friendlyName.length() + 1));
    cardosToken->log(printName);
  }
  catch(CardOSAPI_Exception cardosEx) {
    CardOSError::throwMe(cardosEx);
  }
  
  populate();
}


//
// Database-level ACLs
//
void CardOSToken::getOwner(AclOwnerPrototype &owner)
{
  secdebug("tokend", "CardOSToken::getOwner");
  
  if (cardosToken == NULL) {
    CssmError::throwMe(CSSM_ERRCODE_SELF_CHECK_FAILED);
  }
  
  cardosToken->log("CardOSToken::getOwner");
  
  // we don't really know (right now), so claim we're owned by PIN #0
  if (!mAclOwner) {
    mAclOwner.allocator(Allocator::standard());
    mAclOwner = AclFactory::PinSubject(Allocator::standard(), 0);
  }
  owner = mAclOwner;
}


void CardOSToken::getAcl(const char *tag, uint32 &count, AclEntryInfo *&acls)
{
  secdebug("tokend", "CardOSToken::getAcl");
  
  if (cardosToken == NULL) {
    CssmError::throwMe(CSSM_ERRCODE_SELF_CHECK_FAILED);
  }
  
  // ToDo: Return actuall authentication ids used to protect the token
  //       credentials instead of returning a hard-coded 1.
  
  cardosToken->log("CardOSToken::getAcl");
  if (tag) cardosToken->log(tag);
  
  Allocator &alloc = Allocator::standard();
  
  unsigned pinNum = pinFromAclTag(tag, "?");
  
  if (pinNum > 0) {
    static AutoAclEntryInfoList acl;
    acl.clear();
    acl.allocator(alloc);
	switch (cardosToken->getPinState(pinNum)) {
    case PIN_AUTHENTICATED: 
	  cardosToken->log("CardOSToken::getAcl: CSSM_ACL_PREAUTH_TRACKING_AUTHORIZED");
	  acl.addPinState(pinNum, CSSM_ACL_PREAUTH_TRACKING_AUTHORIZED);
	  break;
	case PIN_BLOCKED:
	  cardosToken->log("CardOSToken::getAcl: CSSM_ACL_PREAUTH_TRACKING_BLOCKED");
	  acl.addPinState(pinNum, CSSM_ACL_PREAUTH_TRACKING_BLOCKED);
	  break;
	case PIN_NOT_AUTHENTICATED:
	  cardosToken->log("CardOSToken::getAcl: CSSM_ACL_PREAUTH_TRACKING_UNKNOWN");
	default:
	  cardosToken->log("CardOSToken::getAcl: CSSM_ACL_PREAUTH_TRACKING_UNKNOWN");
	  acl.addPinState(pinNum, CSSM_ACL_PREAUTH_TRACKING_UNKNOWN);
	  break;
	}
    count = acl.size();
    acls = acl.entries();
    return;
  }
  
  // get pin list, then for each pin
  if (!mAclEntries) {
    mAclEntries.allocator(alloc);
    
    // Anyone can read the attributes and data of any record on this token
    // (it's further limited by the object itself).
    mAclEntries.add(CssmClient::AclFactory::AnySubject( mAclEntries.allocator()),
                    AclAuthorizationSet(CSSM_ACL_AUTHORIZATION_DB_READ, 0));
    
	
	// Get set of all authentication ids used by the token.
	std::set<int> authIds;
	authIds.insert(cardosToken->getUserAuthId()); // Insert default user PIN.
#if 0
	X509CredentialList & creds = cardosToken->getX509Credentials();
	for(X509CredentialList::iterator itr = creds.begin(); itr != creds.end(); itr++) {
	  try {
		authIds.insert((*itr).getAuthId());
	  }
	  catch (CardOSAPI_Exception ex) {}
	}
#endif
	
	// We support all PINs (PIN1, ... PINn) with either a passed in password
	// subject or a prompted password subject.
	for (std::set<int>::iterator itr = authIds.begin(); itr != authIds.end(); itr++) {
	  mAclEntries.addPin(AclFactory::PWSubject(alloc), *itr);
	  mAclEntries.addPin(AclFactory::PromptPWSubject(alloc, CssmData()), *itr);
	}
  }
  count = mAclEntries.size();
  acls = mAclEntries.entries();
}


void CardOSToken::terminate(uint32 reason, uint32 options)
{
  secdebug("tokend", "CardOSToken::terminate");
  
  if (cardosToken == NULL) {
    CssmError::throwMe(CSSM_ERRCODE_SELF_CHECK_FAILED);
  }
     
  cardosToken->log("CardOSToken::terminate");
  
  CardOSAPI_TokenFactory::destroy(cardosToken);
  
  cardosToken = NULL;
}

void CardOSToken::populate()
{
  secdebug("tokend", "CardOSToken::populate begin");
  
  if (cardosToken == NULL) {
    CssmError::throwMe(CSSM_ERRCODE_SELF_CHECK_FAILED);
  }
  
  cardosToken->log("CardOSToken::populate");
  
  // Delete schema if it already exist.
  if (mSchema != NULL) {
    delete mSchema;
  }
  
  // Create new schema
  mSchema = new CardOSSchema();
  mSchema->create();
  
  // Obtain default relations
  Tokend::Relation &certRelation =
  mSchema->findRelation(CSSM_DL_DB_RECORD_X509_CERTIFICATE);
  Tokend::Relation &privateKeyRelation =
  mSchema->findRelation(CSSM_DL_DB_RECORD_PRIVATE_KEY);
  /*
   Tokend::Relation &publicKeyRelation =
   mSchema->findRelation(CSSM_DL_DB_RECORD_PUBLIC_KEY);
   Tokend::Relation &symmetricKeyRelation =
   mSchema->findRelation(CSSM_DL_DB_RECORD_SYMMETRIC_KEY);
   Tokend::Relation &dataRelation =
   mSchema->findRelation(CSSM_DL_DB_RECORD_GENERIC);
  */
  
  // Add credentials
  X509IdentityList & idents = cardosToken->getX509Identities();
  for (X509IdentityList::iterator itr = idents.begin(); itr != idents.end(); itr++) {
    RefPointer<Tokend::Record> cert(new CardOSCertificateRecord(certRelation.metaRecord(), *itr));
    certRelation.insertRecord(cert);
    
    // If there exists a private key for this certificate, we link them together.
    if((*itr).canSign() || (*itr).canUnwrap() || (*itr).canDecrypt()) {
	  RefPointer<Tokend::Record> key(new CardOSKeyRecord(privateKeyRelation.metaRecord(), *itr));
      privateKeyRelation.insertRecord(key);
      
      // Link the key and the certificate.
      key->setAdornment(mSchema->publicKeyHashCoder().certificateKey(), new Tokend::LinkedRecordAdornment(cert));
	}
  }
  
  secdebug("tokend", "CardOSToken::populate end");
}

